#include <iostream>
#include <sys/stat.h>

#include <fastformat/fastformat.hpp>

#include "load_bracketed_dialog.h"
#include "hdrgenerator.h"
#include "openimageform.h"
#include "photoimage.h"
#include "preprocessingoptionsform.h"
#include "../PhotomatixMisc/appproperties.h"
#include "GladeUtils.h"
#include "hdrgenerator.h"
#include "localtext.h"
#include "appproperties.h"


using namespace std;
using namespace PhotomatixImaging;
using namespace PhotomatixMisc;


namespace Photomatix
{
	LoadBracketedDialog::LoadBracketedDialog(BaseObjectType* cobject, const Glib::RefPtr<Gtk::Builder>& builder):
		Gtk::Dialog(cobject),
		m_builder(builder)
	{
		this->set_resizable(false);
		
		Gtk::Button* button;

		// create the file list data model and attach it to the glade tree view
		m_builder->get_widget("filelist-treeview", m_treeview);
		m_filelist = Gtk::ListStore::create(m_columns);
		m_treeview->set_model(m_filelist);
		m_treeview->append_column("File", m_columns.m_col_short_name);
		m_treeview->get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);

		// Set up drag and drop
		std::list<Gtk::TargetEntry> dropTargets;
		dropTargets.push_back( Gtk::TargetEntry("text/uri-list") );

		// Connect Signals
		m_builder->get_widget("cancel-button", button);
		if (button) {
			button->signal_clicked().connect( sigc::mem_fun(*this, &LoadBracketedDialog::on_button_cancel) );
		}

		m_builder->get_widget("ok-button", btnOk);
		if (btnOk) {
			btnOk->signal_clicked().connect( sigc::mem_fun(*this, &LoadBracketedDialog::on_button_ok) ); 
		}
		btnOk->set_sensitive(false);

		m_builder->get_widget("add-button", button);
		if (button) {
			button->signal_clicked().connect( sigc::mem_fun(*this, &LoadBracketedDialog::on_button_add) ); 
		}
		
		m_builder->get_widget("remove-button", button);
		if (button) {
			button->signal_clicked().connect( sigc::mem_fun(*this, &LoadBracketedDialog::on_button_remove) ); 
		}
		
		std::vector<Gtk::TargetEntry> listTargets;
		listTargets.push_back(Gtk::TargetEntry("text/uri-list"));
		drag_dest_set(listTargets, Gtk::DEST_DEFAULT_MOTION | Gtk::DEST_DEFAULT_DROP, Gdk::ACTION_COPY | Gdk::ACTION_MOVE);
		signal_drag_data_received().connect(sigc::mem_fun(*this, &LoadBracketedDialog::on_treeview_dnd_data_received));
		
		this->ReadStickySettings();
	}

	LoadBracketedDialog::~LoadBracketedDialog()
	{
	}
	
	vector <string> LoadBracketedDialog::GetFileNames()
	{
		Gtk::TreeModel::Children rows = m_filelist->children();
		vector <string> filenames;
		for(auto& row : rows)
		{
			Glib::ustring file = row[m_columns.m_col_full_path];
			filenames.push_back(file);
		}
		
		if(filenames.size() == 1)	
		{
			Gtk::MessageDialog dialog(*this, "", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK);
			dialog.set_secondary_text(LocalText::GetTextRequiresAtLeastTwoImages());
			dialog.run();
		}
		
		return filenames;
	}

	void LoadBracketedDialog::on_button_ok()
	{
		this->SaveStickySettings();
		
		//hide();
		this->signal_ok.emit();
	}

	void LoadBracketedDialog::on_button_cancel()
	{
		this->SaveStickySettings();
		
		hide(); //hide() will cause main::run() to end.
	}

	void LoadBracketedDialog::on_button_add()
	{
		//Gtk::FileChooserDialog chooser("Please choose a file", Gtk::FILE_CHOOSER_ACTION_OPEN);
		OpenImageForm openImageDialog;
		openImageDialog.set_transient_for(*this);
		openImageDialog.set_select_multiple(true);

		// show the file selection dialog and block until it returns
		int result = openImageDialog.run();

		// If they clicked OK (instead of cancel or closing the window)
		if (result == Gtk::RESPONSE_OK)
		{
			PhotomatixMisc::AppProperties::SetLastOpenDirectory(openImageDialog.get_current_folder());
			
			std::vector<std::string> filenames = openImageDialog.get_filenames();

			for(std::vector<std::string>::iterator f = filenames.begin(); f != filenames.end(); ++f)
				add_path_to_list(*f);
		}
	}

	void LoadBracketedDialog::on_button_remove()
	{
		std::vector<Gtk::TreeModel::Path> selected = m_treeview->get_selection()->get_selected_rows();
		std::vector<Gtk::TreeModel::Row> todelete;

		/* first form the list of items to delete without modifying the selected list */
		for (auto i = selected.begin(); i != selected.end(); ++i)
		{
			auto iter = m_filelist->get_iter(*i);
			auto row = *iter;
			todelete.push_back(row);
		}

		/* once we have the list of items we can delete them without damaging the first iteration */
		for (std::vector<Gtk::TreeModel::Row>::iterator r = todelete.begin(); r != todelete.end(); ++r)
			m_filelist->erase(**r);
		
		int count = 0;
		Gtk::TreeModel::Children rows = m_filelist->children();
		for(auto& row : rows) ++count;
		if(!count)
			this->btnOk->set_sensitive(false);
	}

	void LoadBracketedDialog::on_treeview_dnd_data_received(
			const Glib::RefPtr<Gdk::DragContext>& context, int, int,
			const Gtk::SelectionData& selection_data, guint, guint time)
	{
		const int length = selection_data.get_length();
		if((length >= 0) && (selection_data.get_format() == 8))
		{
			std::istringstream iss(selection_data.get_data_as_string());
			std::string uri;
			
			while (std::getline(iss, uri)) {
				if (uri.substr(0,8) == "file:///") {
					std::string full_path = Glib::filename_from_uri(uri);

					// chop off any newlines and carriage returns in the uri list
					while (full_path.c_str()[full_path.size()-1] == '\n' || full_path.c_str()[full_path.size()-1] == '\r') {
						full_path.erase(full_path.size()-1);
					}

					if(PhotomatixImaging::PhotoImage::GetProbableFileFormat(full_path) != PhotomatixImaging::EImageFormatType::Unknown)
						this->add_path_to_list(full_path);
					else
					{
						string message;
						fastformat::fmt(message, "File '{0}' not added, this type of file is not allowed.", full_path.substr(full_path.find_last_of("/") + 1));
						
						Gtk::MessageDialog dialog(*this, "", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK);
						dialog.set_secondary_text(message);
						dialog.run();
					}
				}
			}
			
		}

		context->drag_finish(false, false, time);
	}

	void LoadBracketedDialog::add_path_to_list(const std::string& full_path) 
	{
		/* check if the file exists first */
		struct stat buf;
		if (stat(full_path.c_str(), &buf) == 0) {

			btnOk->set_sensitive(true);
			
			const std::string short_name = full_path.substr(full_path.find_last_of("/") + 1);

			Gtk::TreeModel::Children rows = m_filelist->children();
			bool contains = false;
			for(auto& row : rows)
			{
				if(row[m_columns.m_col_short_name] == short_name)
				{
					contains = true;
					break;
				}
			}
			
			if(!contains)
			{
				Gtk::TreeModel::Row row = *(m_filelist->append());
				row[m_columns.m_col_short_name] = short_name;
				row[m_columns.m_col_full_path] = full_path;
			}
			else
			{			
				string message;
				fastformat::fmt(message, "File '{0}' is already in the list.", short_name);
				Gtk::MessageDialog dialog(*this, "", false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK);
				dialog.set_secondary_text(message);
				dialog.run();
			}
		}
	}
	
	void LoadBracketedDialog::ReadStickySettings()
	{
	}

	void LoadBracketedDialog::SaveStickySettings()
	{
	}

	bool LoadBracketedDialog::on_window_delete(GdkEventAny *event)
	{
		hide(); //hide() will cause main::run() to end.
		return true;
	}

}