#include "processphotocontroller.h"
#include "displayableimageform.h"

#include "previewcontroller.h"
#include "hdrcontroller.h"
#include "finishingtouchcontroller.h"
#include "finishingtouch.h"
#include "../PhotomatixImaging/hdrgenerator.h"
#include "../PhotomatixImaging/photoimage.h"
#include "../PhotomatixMisc/localtext.h"
#include "../PhotomatixMisc/appproperties.h"
#include "../Photomatix32/pmjp/pmjp.h"
#include "../Photomatix32/PhotomatixMetaLib.h"
#include "../Photomatix32/exiv2Wrapper.h"
#include "../PhotomatixMisc/GladeUtils.h"

#include <cassert>


using namespace PhotomatixImaging;
using namespace PhotomatixMisc;
using namespace std;


namespace Photomatix
{
	ProcessPhotoController::ProcessPhotoController():
		window(NULL)
	{
		this->finalImage = nullptr;
	}

	ProcessPhotoController::~ProcessPhotoController()
	{

	}
	
	void ProcessPhotoController::ShowImageWindow()
	{
		this->CreateImageWindow();
		
		assert(this->window);
		this->window->present();
	}
	
	void ProcessPhotoController::CreateImageWindow()
	{
		if(!this->window)
		{
			PhotomatixUtils::get_widget_derived("displayable-image-dialog", this->window);
			this->window->signal_file_open_requested.connect(sigc::mem_fun(*this, &ProcessPhotoController::OnCreateInstanceRequested));
			this->window->signal_tonemapping_requested.connect(sigc::mem_fun(*this, &ProcessPhotoController::OnToneMappingRequested));
			this->window->signal_save_image_requested.connect(sigc::mem_fun(*this, &ProcessPhotoController::OnSaveImageRequested));
		}
	}
	
	void ProcessPhotoController::ImageWindowOpenFile(const string& fileName)
	{
		assert(this->window);
		// Opening image was cancelled or there was an error
		if(!this->window->OpenImage(fileName))
			return;
		// If RAW file is opened we don't display it but instead go directly into preview mode
		if(PhotoImage::GetProbableFileFormat(fileName) == EImageFormatType::Raw || PhotoImage::GetProbableFileFormat(fileName) == EImageFormatType::Hdr)
		{	
			OnToneMappingRequested();
		}
	}
	
	void ProcessPhotoController::OnToneMappingRequested()
	{
		// Prompt for noise reduction in case of single LDR files
		if(!this->window->GetHdrController() && (this->window->GetPhoto()->GetImageFormat() == EImageFormatType::Jpeg ||
			this->window->GetPhoto()->GetImageFormat() == EImageFormatType::Tiff) && AppProperties::GetReduceNoiseLdr())
		{
			Gtk::MessageDialog dialog(*this->window, "Noise reduction", false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO);
			dialog.set_secondary_text(LocalText::GetTextReduceNoiseBeforeToneMapping());
			int result = dialog.run();
			if(result == Gtk::RESPONSE_YES) {
				this->window->GetPhoto()->ReduceNoise(100);
			}
		}
		
		this->previewController = shared_ptr <PreviewController>(new PreviewController(this->window->GetPhoto(), this->window->GetHdrController() ? this->window->GetHdrController()->GetHdrGenerator() : NULL));
		this->previewController->signal_preview_closed.connect(sigc::mem_fun(*this, &ProcessPhotoController::OnPreviewClosed));
		this->previewController->signal_preview_finished.connect(sigc::mem_fun(*this, &ProcessPhotoController::OnPreviewFinished));
		this->window->hide();
	}
	
	void ProcessPhotoController::OnPreviewClosed()
	{
		this->window->show();
		this->previewController.reset();
	}
	
	void ProcessPhotoController::OnPreviewFinished(shared_ptr<PhotoImage> outputImage)
	{
		// Show Finishing Touch, if needed. Save the image, otherwise
		if(AppProperties::GetShowFinishingTouch()) {
			
			//this->previewController.reset();
			
			this->window->SetMenuItemsEnability(false);
	
			this->finishingTouchController = shared_ptr <FinishingTouchController>(new FinishingTouchController(outputImage, this->window));
			this->finishingTouchController->signal_finishingtouch_finished.connect(sigc::mem_fun(*this, &ProcessPhotoController::OnFinishingTouchFinished));
			this->finishingTouchController->signal_finishingtouch_closed.connect(sigc::mem_fun(*this, &ProcessPhotoController::OnFinishingTouchClosed));
			this->finishingTouchController->signal_finishingtouch_preview_updated.connect(sigc::mem_fun(*this, &ProcessPhotoController::OnFinishingTouchPreviewUpdated));
			this->finishingTouchController->Init();
			
			this->window->signal_mouse_moved.connect(sigc::mem_fun(*finishingTouchController, &FinishingTouchController::OnImageMouseMoved));
			this->window->signal_mouse_down.connect(sigc::mem_fun(*finishingTouchController, &FinishingTouchController::OnImageMouseDown));
			this->window->signal_mouse_up.connect(sigc::mem_fun(*finishingTouchController, &FinishingTouchController::OnImageMouseUp));
			
		} else {
			this->SaveFinalImage(outputImage);
		}
	}
	
	void ProcessPhotoController::OnFinishingTouchClosed(shared_ptr< PhotoImage > outputImage)
	{
		this->finalImage = outputImage;
		this->window->SetMenuItemsEnability(false);
	}
	
	void ProcessPhotoController::OnFinishingTouchFinished(shared_ptr<PhotoImage> outputImage)
	{
	//	this->window->SetPhoto(outputImage);
		this->SaveFinalImage(outputImage);
		this->window->SetMenuItemsEnability(false);
	}

	void ProcessPhotoController::OnFinishingTouchPreviewUpdated(shared_ptr<PhotoImage> previewImage)
	{
		this->window->SetPhoto(previewImage, false);
		this->window->show();
		this->window->SetMenuItemsEnability(false);
	}
	
	void ProcessPhotoController::OnSaveImageRequested()
	{
		if(this->finalImage) {
			this->SaveFinalImage(this->finalImage);
		}
	}
	
	void ProcessPhotoController::SaveFinalImage(shared_ptr<PhotoImage> outputImage)
	{
		// Save the image
		
		shared_ptr <HdrGenerator> hdrGenerator = this->previewController->GetHdrGenerator();
		shared_ptr <PhotoImage> originalPhoto = this->previewController->GetOriginalPhoto();
		
		string metaSrcImage = hdrGenerator ? hdrGenerator->GetMedianPhoto()->GetFileName() : originalPhoto->GetFileName();
		
		string prevFolder = saveFileDialog.get_current_folder();
		string fileName = hdrGenerator ? hdrGenerator->GetOutputFileName(AppProperties::GetConcatenateSourceFilenames(), false) : originalPhoto->GetFileNameShortNoExtension();
		// Add suffix
		if(AppProperties::GetAppendSuffix()) {
			if(AppProperties::GetAutomaticSuffix())
			{
				fileName = fileName + (this->previewController->GetCurrentMethod() == 10 ? "_fused" : "_tonemapped");
			}
			else
				fileName = fileName + AppProperties::GetCustomSuffix();
		}
		if(saveFileDialog.get_filter()->get_name() == "JPEG")
			fileName = fileName + ".jpg";
		else 
			fileName = fileName + ".tif";
		
		//if(AppProperties::GetUseLastOpenDirForSave())
		if(!AppProperties::GetLastSaveDirectory().empty())
			saveFileDialog.set_current_folder(PhotomatixMisc::AppProperties::GetLastSaveDirectory());
		/*else {
			shared_ptr <PhotoImage> photo = hdrGenerator ? hdrGenerator->GetInputPhoto(0) : originalPhoto;
			string directory = photo->GetFileName();
			int index = directory.find_last_of("//");
			directory = directory.substr(0, index);
			saveFileDialog.set_current_folder(directory);
		}*/
		saveFileDialog.set_current_name(fileName);
		//if(saveFileDialog.get_current_folder().empty())
		//	saveFileDialog.set_current_folder(prevFolder);
		saveFileDialog.set_do_overwrite_confirmation(true);
		saveFileDialog.property_filter().signal_changed().connect(sigc::mem_fun(*this, &ProcessPhotoController::OnSaveDialogFilterChanged));
		int saveResult = saveFileDialog.run();
		saveFileDialog.set_visible(false);
		
		// User cancelled so return
		if(saveResult == Gtk::RESPONSE_CANCEL || saveResult == -4) {
			this->window->SetSaveMenuEnability(true);
			this->finalImage = outputImage;
			return;
		}
		
		this->finalImage = nullptr;

		
		PhotomatixMisc::AppProperties::SetLastSaveDirectory(saveFileDialog.get_current_folder());
		
		bool saveAs16Bit = false;
		if(saveFileDialog.get_filter()->get_name() == "JPEG") {
			AppProperties::SetSaveDialogFilter(0);
		} else if(saveFileDialog.get_filter()->get_name() == "TIFF 16-bit") {
			AppProperties::SetSaveDialogFilter(1);
			saveAs16Bit = true;
		} else if(saveFileDialog.get_filter()->get_name() == "TIFF 8-bit") {
			AppProperties::SetSaveDialogFilter(2);
		}
		
		if(!saveAs16Bit) {
			outputImage->ReduceBitDepthFrom16to8();
		}
		
		
		stringstream path;
		string name = saveFileDialog.get_filename();
		// Strip extension from file name
		size_t index = name.find_last_of('.');
		if(index != string::npos)
			name = name.substr(0, index);
		path << name;
		// Add suffix
		if(AppProperties::GetAppendSuffix()) {
			if(AppProperties::GetAutomaticSuffix())
			{
				path << (this->previewController->GetCurrentMethod() == 10 ? "_fused" : "_tonemapped");
			}
			else
				path << AppProperties::GetCustomSuffix();
		}
		path << "." << (saveFileDialog.get_filter()->get_name() == "JPEG" ? "jpg" : "tif");
		
		// Apply watermark if necessary
		if(!AppProperties::IsRegistered()) {
			outputImage->ApplyWatermark();
		}
		
		int result = 0;
		if(saveFileDialog.get_filter()->get_name() == "JPEG")
		{
			vector <int> saveInfo = outputImage->GetSaveInfoBuffer();
			//if(this->currentMethod < 10) saveInfo[7] = 1;
			result = PMwritejpg(path.str().c_str(), AppProperties::GetJpegQuality(), saveInfo.data(), outputImage->GetRowStride(),
				outputImage->GetIccProfile().data(), outputImage->GetExifBuffer().data(), outputImage->GetGpsBuffer().data(),
				outputImage->GetIptcBuffer().size(), outputImage->GetIptcBuffer().data(), this->previewController->GetCurrentMethod(),
				this->previewController->GetCurrentSettings(), outputImage->GetImageBuffer()->data());
		}
		else
		{
			// TODO: 0 and NULL at the end of below call should be replaced with this.xmpLen and this.xmpData respectively
			result = buf2tif(outputImage->GetImageBuffer()->data(), outputImage->GetSaveInfoBuffer().data(), outputImage->GetRowStride() * (outputImage->GetBitsPerSample() >> 3),
				outputImage->GetIccProfile().data(), outputImage->GetExifBuffer().data(), outputImage->GetGpsBuffer().data(),
				outputImage->GetIptcBuffer().size(), outputImage->GetIptcBuffer().data(), this->previewController->GetCurrentMethod(), this->previewController->GetCurrentSettings(), 0, NULL, path.str().c_str());
		}
		
		// Write EXIF using exifTool
		char softwareTag[200];
		snprintf(softwareTag, sizeof(softwareTag), "%s",  "Photomatix for Linux 1.1");
		char exifToolPath[MAX_PATH];
		char* tempIn = tempnam("/tmp/", "cl1");
		char* tempOut = tempnam("/tmp/", "cl2");
		strcpy(exifToolPath, "/usr/bin/exiftool");
		result = wrapMetadata(metaSrcImage.c_str(), path.str().c_str(), exifToolPath, tempIn, tempOut, softwareTag, 0);
		remove(tempIn);
		remove(tempOut);
		
		
		// Create and show new DisplayableImageForm
		this->window->SetUnsavedChanges(false);
		this->window->Close();
		if(DisplayableImageForm::OpenedImagesCount == 0)
		{
			this->ShowImageWindow();
		}
		
		this->previewController.reset();
	}

	void ProcessPhotoController::OnSaveDialogFilterChanged()
	{
		string fileName = saveFileDialog.get_filename();//saveFileDialog.get_current_name();
		int index_file = fileName.find_last_of('/');
		int index_ext = fileName.find_last_of('.');
		fileName = fileName.substr(index_file + 1, index_ext - index_file - 1);
		
		if(saveFileDialog.get_filter()->get_name() == "JPEG")
			fileName = fileName + ".jpg";
		else 
			fileName = fileName + ".tif";
		saveFileDialog.set_current_name(fileName);
	}

	void ProcessPhotoController::OnCreateInstanceRequested(string fileName)
	{
		if(DisplayableImageForm::OpenedImagesCount == 0)
		{
			this->ShowImageWindow();
			this->ImageWindowOpenFile(fileName);
		}
		else
		{
			ProcessPhotoController* newController = new ProcessPhotoController();
			newController->ShowImageWindow();
			newController->ImageWindowOpenFile(fileName);
		}
		
		++DisplayableImageForm::OpenedImagesCount;
	}

}