#ifndef DISPLAYABLEIMAGE_H
#define DISPLAYABLEIMAGE_H

#include <gtkmm/enums.h>
#include <gtkmm/container.h>
#include <gtkmm/image.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/box.h>
#include <gtkmm/drawingarea.h>
#include <memory>

namespace PhotomatixImaging
{
	class PhotoImage;
}

namespace PhotomatixControls
{
	/** Control which uses Cairo rendering and transformatio to draw pixel buffer.
	 * @remarks
	 * 	Using Cairo results in a much better performance than when using regular Gdk calls.
	 */
	class DisplayableDrawingArea : public Gtk::DrawingArea
	{
	protected:
		Glib::RefPtr<Gdk::Pixbuf> pixbuf;						///< Pixel buffer with image data
		
		int offsetX;
		int offsetY;
		float scale;
		
		Gdk::Rectangle cropRect;
		
		bool hasBorder;
		
		/** All the rendering happens here.
		 * 	@par
		 * 		Pixel buffer is transformed by transformation matrix containing
		 * 		translation (offsetX and offsetY member variables) and scaling
		 * 		(scale member variable).
		 */
		virtual bool on_draw(const Cairo::RefPtr<Cairo::Context>& cr);
		
		/// Emits signal_clicked event.
		virtual bool on_button_press_event(GdkEventButton* event)
		{
			/// TODO: maybe only pressing left-button should result in this behaviour?
			this->signal_mouse_down.emit(event);
			
			return true;
		}
		
		virtual bool on_motion_notify_event (GdkEventMotion* motion_event)
		{
			this->signal_mouse_moved.emit(motion_event);
			
			return true;
		}
		
		virtual bool on_button_release_event(GdkEventButton* release_event)
		{
			this->signal_mouse_up.emit(release_event);
			
			return true;
		}
		
	public:
		DisplayableDrawingArea():
			cropRect(0, 0, 0, 0)
		{
			this->scale = 1.0f;
			this->offsetX = 0.0f;
			this->offsetY = 0.0f;
			this->hasBorder = false;
		}
		
		void SetPixBuf(Glib::RefPtr<Gdk::Pixbuf> pixbuf) {this->pixbuf = pixbuf; this->queue_draw();}
		void SetOffset(int offsetX, int offsetY) {this->offsetX = offsetX; this->offsetY = offsetY;}
		void SetScale(float scale) {this->scale = scale;}
		void ShowCropRectangle(Gdk::Rectangle cropRect);
		void SetBorder(bool border);
		
		sigc::signal <void, GdkEventButton*> signal_mouse_down;
		sigc::signal <void, GdkEventButton*> signal_mouse_up;
		sigc::signal <void, GdkEventMotion*> signal_mouse_moved;
	};
	
	
	/** Control displaying a PhotoImage object.
	*/
	class DisplayableImage
	{
	public:
		DisplayableImage();
		DisplayableImage(const std::string& fileName);
		virtual ~DisplayableImage();
		
		/// Returns PhotoImage object associated with this control.
		std::shared_ptr <PhotomatixImaging::PhotoImage> GetPhoto() {return this->photo;}
		
		/** Creates new PhotoImage object by loading it from file. Then this image is displayed
		 * by the control.
		 */
		virtual void Open(const std::string& fileName, bool openRAWs=false, int colorTemperature=0, int colorSpace=0);
		
		/// Method that allows showing existing PhotoImage object.
		virtual void SetPhoto(std::shared_ptr <PhotomatixImaging::PhotoImage> photo);
		
		/// Sets new zoom factor. This controls magnification of the photo.
		virtual void SetZoomFactor(float zoomFactor, int maxWidth=0, int maxHeight=0) {}
		
		/// Increases exposure of HDR image.
		void ExposureUp();
		/// Decreases exposure of HDR image.
		void ExposureDown();
		
		virtual void RemoveChildControl() {}
		void ShowCropRectangle(Gdk::Rectangle cropRect);
		void SetBorder(bool hasBorder);
		
		
		sigc::signal <void, GdkEventButton*> signal_mouse_down;
		sigc::signal <void, GdkEventButton*> signal_mouse_up;
		sigc::signal <void, GdkEventMotion*> signal_mouse_moved;
		
	private:
		float exposureLevel;							///< Exposure level to preview HDR images		
		
		void UpdatePixBuf();
		
	protected:
		
		DisplayableDrawingArea* image;						///< DrawingArea control used to show a photo
		
		Glib::RefPtr <Gdk::Pixbuf> pixBuf;					///< Copy of the photo image buffer
		std::shared_ptr <PhotomatixImaging::PhotoImage> photo;			///< PhotoImage associated with this control
		bool setUp;								///< Is this control already initialized?
		
		
		//@{
		/** Some helper variables */
		float previousZoom;
		int previousWidth;
		int previousHeight;
		//@}
		
		void OnMouseDown(GdkEventButton* evt);
		void OnMouseUp(GdkEventButton* evt);
		void OnMouseMoved(GdkEventMotion* evt);
	};
}

#endif //GTKMM_CUSTOM_CONTAINER_MYCONTAINER_H
