#ifndef PHOTOIMAGE_H
#define PHOTOIMAGE_H

#include <memory>
#include <string>
#include <vector>
#include <sstream>
#include <iostream>

namespace PhotomatixImaging
{
	/// Format of the image file
	enum EImageFormatType : unsigned int
	{
		Unknown=0,
		Jpeg=1,
		Tiff=2,
		//Tif32,
		Exr=4,							///< Initially not supported
		Hdr=5,
		Raw=6,
		Psd=7
	};
	
	// Color profile
	enum EColorProfile
	{
		sRgb = 0,
		AdobeRgb = 1,
		ProPhotoRgb = 2
	};

	/** Class holding image data.
	*/
	class PhotoImage
	{
	private:
		unsigned int width;							///< Width of the photo
		unsigned int height;						///< Height of the photo
		unsigned int samplesPerPixel;
		unsigned int bitsPerSample;					///< Bits per each sample of the photo
		unsigned int flip;
		unsigned int orientation;
		
		int colorTemperature;
		int colorSpace;
		int DPI;
		int processingMethod;
		
		int globalImageBufferIndex;					///< Index of the image in the global image buffer (-1 by default)
		
		EImageFormatType imageFormat;				///< Image format of the photo associated with this PhotoImage object
		bool isZipCompressed;
		
		/// Image buffer holding pixels information. Used in case of sinle photos.
		std::shared_ptr <std::vector<unsigned char>> imageBuffer;
		/** Global image buffer. It contains data for all images and is used only in case of bracketed sets
		 */
		std::shared_ptr <std::vector<unsigned char>> globalImageBuffer;
		
		std::vector <int> infoBuffer;
		std::vector <int> exifBuffer;
		std::vector <int> gpsBuffer;
		std::vector <float> rgbwCoords;

		std::vector <char> xmpBuffer;
		std::vector <char> lensBuffer;
		std::vector <unsigned char> iccProfile;
		std::vector <float> pixelDimensions;
		
		std::vector <unsigned char> iptcBuffer;
		
		/// Name of the file on disk.
		std::string photoFileName;
		
		void RemoveBuffer();		
		
		/// Creates vector of unsigned chars to hold image data.
		std::shared_ptr <std::vector <unsigned char>> AllocateImageBuffer(bool isBatch, int width, int height, int rowStride, int samplesPerPixel, int bytesCount);
		
		/// Retrives file information such as image dimensions, samples per pixels etc.
		void SetFileInfo();
		void SetIptcInfo(int iptcOffset, int iptcLen);
		/// Reads Icc profile from file.
		void SetIccProfile();
		
		//@{
		/** Functions to load image data from various file formats */
		void ImportFromJpeg(bool isBatch);
		void ImportFromTiff(bool isBatch);
		void ImportFromRaw(bool isBatch);
		void ImportFromPsd(bool isBatch);
		void ImportFromHdr();
		
		void ReadJpegFile(bool isBatch, long offset=0);
		void ReadTiffFile(bool isBatch, long offset=0);
		void ReadPsdFile(bool isBatch, long offset=0);
		//@}
		
		/// Reads EXIF data from file.
		void SetExifData();
		/// Reads XMP data from file.
		void SetXMPData();


		void ReduceNoiseHDR(int strength);
		void ReduceNoiseLDR(float strengthFactor);
		
		// T must be an iterable container
		template <typename T> void exportContainer(std::ostringstream &os, const T& cont) const
		{
		  int len = cont.size();
		  os.write(reinterpret_cast<char*>(&len), sizeof(len));
		  for (auto &elt : cont)
		    os.write(reinterpret_cast<const char*>(&elt), sizeof(elt));
		}
		
		// T must be a container with T(int size), &T[0] and elements sequential in memory
		// T::value_type must be default constructable
		template <typename T> T importContainer(std::istringstream &is) const
		{
		  int len;
		  is.read(reinterpret_cast<char*>(&len), sizeof(len));
		  T cont(len);
		  is.read(reinterpret_cast<char*>(&cont[0]), len * sizeof(typename T::value_type));
		  return cont;
		}

	public:
		PhotoImage();
		explicit PhotoImage(const std::string& photoFileName, bool loadBuffer);
		explicit PhotoImage(const std::string& photoFileName, bool loadBuffer, int colorSpace, int colorTemperature);
		explicit PhotoImage(unsigned int width, unsigned int height, unsigned int samplesPerPixel, unsigned int bitsPerSample, EImageFormatType imageFormat);
		~PhotoImage();
		
		/// Loads image buffer from external files.
		void LoadBuffer(bool isBatch=false, unsigned int imagesCount=1);
		
		/// Returns image buffer associated with this PhotoImage.
		std::shared_ptr <std::vector<unsigned char>> GetImageBuffer() const {return this->imageBuffer;}
		void SetImageBuffer(std::shared_ptr<std::vector<unsigned char>> buffer) {this->imageBuffer = buffer;}
		/// Returns true if PhotoImage has buffer assigned.
		bool HasBuffer() const {return (this->imageBuffer != 0 && this->imageBuffer->size() > 0);}
		
		void SetGlobalImageBuffer(std::shared_ptr <std::vector<unsigned char>> buffer) {this->globalImageBuffer = buffer;}
		std::shared_ptr <std::vector<unsigned char>> GetGlobalImageBuffer() {return this->globalImageBuffer;}
		
		void SetWidth(unsigned int width) {this->width = width;}
		unsigned int GetWidth() const {return this->width;}
		
		void SetHeight(unsigned int height) {this->height = height;}
		unsigned int GetHeight() const {return this->height;}
		
		unsigned int GetXResolution() const {return this->infoBuffer[4];}
		unsigned int GetYResolution() const {return this->infoBuffer[5];}
		
		unsigned int GetSamplesPerPixel() const {return this->samplesPerPixel;}
		unsigned int GetBitsPerSample() const {return this->bitsPerSample;}
		void SetBitsPerSample(unsigned int bitsPerSample) {this->bitsPerSample = bitsPerSample;}

		std::vector <float> GetRgbwCoords() const {return this->rgbwCoords;}
		
		unsigned int GetOrientation() const {return this->orientation;}
		void SetOrientation(unsigned int orientation) {this->orientation = orientation;}
		
		void SetGlobalBufferIndex(unsigned int index) {this->globalImageBufferIndex = index;}
		unsigned int GetGlobalBufferIndex() {return this->globalImageBufferIndex;}
		
		void SetPixelDimensions(const std::vector <float>& pixelDims) {this->pixelDimensions = pixelDims;}
		void SetHDRPixelDimensions();

		bool IsZipCompressed() const {return this->isZipCompressed;}
		
		float GetPixelAverage() const {return this->pixelDimensions[2];}
		
		size_t GetBoundedImageBufferSize(bool isBatch) const
		{
			// TODO: rowStride is not really supported at the moment
			return (isBatch ? (size_t)this->width * this->height * this->samplesPerPixel * this->GetBytesPerPixel() : 0);// : this->rowStride * this->height * this->GetBytesPerPixel();
		}
		
		bool IsHDR() const {return this->imageFormat == EImageFormatType::Hdr;}
		
		/// Returns number of bytes per pixel.
		unsigned int GetBytesPerPixel() const {return this->bitsPerSample >> 3;}
		
		void SetFlip(unsigned int flip) {this->flip = flip;}
		unsigned int GetFlip() const {return this->flip;}
		
		/// Sets EXIF data of this photo.
		void SetExifBuffer(const std::vector<int>& exifBuffer) {this->exifBuffer = exifBuffer;}
		/// Returns reference to EXIF data of this photo.
		std::vector<int>& GetExifBuffer() {return this->exifBuffer;}
		
		/// Sets GPS data of this photo.
		void SetGpsBuffer(const std::vector<int>& gpsBuffer) {this->gpsBuffer = gpsBuffer;}
		/// Returns reference to GPS data of this photo.
		std::vector<int>& GetGpsBuffer() {return this->gpsBuffer;}
		
		/// Sets IPTC data of this photo.
		void SetIptcBuffer(const std::vector<unsigned char>& iptcBuffer) {this->iptcBuffer = iptcBuffer;}
		/// Returns reference to IPTC data of this photo.
		std::vector<unsigned char>& GetIptcBuffer() {return this->iptcBuffer;}
		
		/// Sets XMP buffer.
		void SetXmpBuffer(const std::vector<char>& xmpBuffer) {this->xmpBuffer = xmpBuffer;}
		/// Returns reference to XMP buffer.
		std::vector<char>& GetXMPBuffer() {return this->xmpBuffer;}
		
		/// Sets lens buffer.
		void SetLensBuffer(const std::vector<char>& lensBuffer) {this->lensBuffer = lensBuffer;}
		/// Returns reference to lens buffer.
		std::vector<char>& GetLensBuffer() {return this->lensBuffer;}
		
		/// Sets ICC profile of this photo.
		void SetIccProfile(const std::vector<unsigned char>& iccProfile) {this->iccProfile = iccProfile;}
		/// Returns ICC profile of this photo.
		std::vector<unsigned char>& GetIccProfile() {return this->iccProfile;}
		
		/// Returns image format of this photo.
		EImageFormatType GetImageFormat() {return this->imageFormat;}
		/// Sets image format of this photo - might be useful when creating new PhotoImage from in-memory data.
		void SetImageFormat(EImageFormatType format) {this->imageFormat = format;}
		
		void SetSize(unsigned int width, unsigned int height) {this->width = width; this->height = height;}
		
		unsigned int GetRowStride() const;
		
		/// Returns DPI of this photo.
		unsigned int GetDPI() const {return this->DPI;}
		/// Sets DPI of this photo.
		void SetDPI(unsigned int dpi) {this->DPI = dpi;}
		
		/// Returns file name of the photo associated with this object.
		std::string& GetFileName() {return this->photoFileName;}
		/// Returns file name of the photo without file path.
		std::string GetFileNameShort() const;
		/// Returns file name of the photo without file path and without extension.
		std::string GetFileNameShortNoExtension() const;
		
		void SetColorTemperature(int temperature) {this->colorTemperature = temperature;}
		void SetColorSpace(int colorSpace) {this->colorSpace = colorSpace;}
		void SetFileName(const std::string& fileName) {this->photoFileName = fileName;}

		/// Reduces Chromatic Aberrations.
		int ReduceChromaticAberrations(int realBps);
		
		void ApplyWatermark();

		std::shared_ptr <PhotoImage> GetThumbnailImage();
		
		/** Returns probable file format based on the image extension
			@remarks
				This method is not guaranteed to return 'real' image file format
				as it guesses it based on the file name.
			@return
				EImageFormatType probable file format. Unknown is returned if none of the formats matches file extension.
		*/
		static EImageFormatType GetProbableFileFormat(const std::string& photoFileName);

		/// Returns row stride for an image with given width and samples per pixel.
		static int CalculateRowStride(int width, int samplesPerPixel);

		/// Returns proportional size for an image to fit in given size.
		void GetProportionalSize(int fitInWidth, int fitInHeight, int& outWidth, int& outHeight);
		
		/// Returns proportional size for an image to fit in given size.
		static void GetProportionalSize(int fitInWidth, int fitInHeight, int inWidth, int inHeight, int& outWidth, int& outHeight);

		/// Reduces noise in this image.
		void ReduceNoise(int noiseStrength=100);

		/// Returns data necessary to save the image.
		std::vector <int> GetSaveInfoBuffer();

		// Returns image metadata as binary data, ready to be attached to an image
		std::vector<char> GetBinaryMetadata();
		
		// Returns a downsampled version of the image
		std::shared_ptr<PhotoImage> GetDownsampledImage(int newWidth, int newHeight, int newBps = -1);
		
		
		void ReduceBitDepthFrom16to8();
		
	public:
		std::string GetSerializedMetadata() const;

		void RestoreSerializedMetadata(const std::string &data);
	};
}
	
#endif // PHOTOIMAGE_H
