#ifndef PHOTOPROCESSOR_H
#define PHOTOPROCESSOR_H


#include <memory>
#include <vector>
#include <string>

#ifndef WIN32
	#include "workerthread.h"
#endif


#if defined(WIN32) && !defined(_CONSOLE)
	class Win32ProgressNotifier;
#endif


namespace PhotomatixImaging
{
	// Forward declarations
#ifndef WIN32
	enum EImageFormatType : unsigned int;
#else
	enum EImageFormatType;
#endif
	class PhotoImage;
	class ProgressNotifier;
	
	
	/// Type of noise reduction
	enum ENoiseReductionType
	{
		NoiseReductionNone = -1,
		NoiseReductionUnderExposed = 0,				/// Noise reduction will be applied only on under exposed images
		NoiseReductionNormalAndUnderExposed = 1,	/// Noise reduction will be applied only on under exposed and normal exposure images
		NoiseReductionAll = 2,						/// Noise reduction will be applied to all images
	};
	
	/// Type of aligment
	enum EAlignmentType
	{
		AlignmentNone = -1,
		AlignmentTranslation = 0,
		AlignmentRotationPerspective = 1,
		AlignmentRotation = 2
	};
	
	/// Ghosting detection strength
	//enum EGhostingDetectionStrength
	//{
	//	DetectionNormal = 0,
	//	DetectionHigh = 1
	//};
	
	
	/** Base photo processor class.
	 */
#ifndef WIN32
	class PhotoProcessor: public PhotomatixMisc::WorkerThread
#else
	class PhotoProcessor
#endif
	{
	protected:
		unsigned int width;							///< Width of the photo
		unsigned int height;						///< Height of the photo
		unsigned int downsampleWidth;
		unsigned int downsampleHeight;
		unsigned int samplesPerPixel;
		unsigned int bitsPerSample;					///< Number of bits per sample
		unsigned int bitsPerSampleOut;				///< Number of bits per sample (of the output image)
		size_t size;								///< Size of a single photo processed by this processor
		
		unsigned int flip;
		unsigned int orientation;
		
		unsigned int colorTemperature;
		unsigned int colorSpace;
		unsigned int rowStride;
		unsigned int dpi;
		
		unsigned int noiseReductionStrength;		/// Strength of the noise reduction
		
		float maxShiftPerspective;
		
		EImageFormatType imageFormat;
		ENoiseReductionType noiseReductionType;
		EAlignmentType alignmentType;
		
		std::vector <int> exifBuffer;
		std::vector <int> gpsBuffer;
		std::vector <unsigned char> iptcBuffer;
		std::vector <char> xmpBuffer;
		std::vector <char> lensBuffer;
		std::vector <unsigned char> iccProfile;
		
		bool isIntegre;
		bool doLoadImages;							///< Is it necessary to load images?
		bool alignImages;							///< Is it necessary to align images?
		bool cropImages;							///< Is cropping images after aligment enabled?
		bool bytesUnaligned;
		bool isPreview;								///< Is generating preview or final image

#ifdef WIN32
		volatile bool stop;
#endif
		
		
		std::shared_ptr <std::vector<unsigned char>> imageBuffer;
		
		/// Vector of input photos
		std::vector <std::shared_ptr <PhotoImage>> inputPhotos;
		/// Result of the processing
		std::shared_ptr <PhotoImage> outputPhoto;
		
		
		virtual void Process();
		/// Processes photo.
		virtual void ProcessPhoto()=0;
		/// Sets processor properties (like width and bits per sample) based on the input images.
		void SetImageProperties();
		
		/// Loads images from files.
		void LoadImages();
		/// Reduces noise on the images.
		void ReduceNoise();
		/// Attempts to align images.
		void AlignImages();
		
		int GetNumberOfImagesToDenoise();
		void GetDelta(std::vector <float>& delta);
		
		void GetInputPhotoBounds(int& minWidth, int& minHeight, int& maxWidth, int& maxHeight);
		
		int GetDocType() const {return (this->bitsPerSample == 32 ? 0 : (this->bitsPerSample == 16 ? 1 : 2));}
		
	public:
		PhotoProcessor();
		virtual ~PhotoProcessor();
		
		/// Loads given photos from disk to the memory
		void LoadInputPhotos(const std::vector <std::string>& fileNames);
		/** Checks integrity of the photos.
		 * @par
		 * 	This includes:
		 * 	1) Checking if bits per sample is the same for all photos
		 *  2) Checking if samples per pixel is the same for all photos
		 *  3) Checking if image format is the same for all photos
		 */
		bool CheckPhotosIntegrity();
		
		unsigned int GetWidth() const {return this->width;}
		void SetWidth(int value);
		unsigned int GetHeight() const {return this->height;}
		void SetHeight(int value) {this->height = value;}
		unsigned int GetSamplesPerPixel() const {return this->samplesPerPixel;}
		void SetSamplesPerPixel(int value) {this->samplesPerPixel = value;}
		unsigned int GetBitsPerSample() const {return this->bitsPerSample;}
		void SetBitsPerSample(int value) {this->bitsPerSample = value;}
		size_t GetSize() const {return this->size;}
		void SetSize(size_t value) {this->size = value;}

		std::shared_ptr <std::vector<unsigned char>> GetImagesBuffer() {return this->imageBuffer;}
		void SetImagesBuffer(std::shared_ptr <std::vector<unsigned char>> buffer) {this->imageBuffer = buffer;}
		
		void SetInputPhotos(std::vector <std::shared_ptr <PhotoImage>> inputPhotos) {this->inputPhotos = inputPhotos; this->SetImageProperties();}

		/// Returns number of input photos.
		unsigned int GetInputPhotosCount() const {return this->inputPhotos.size();}
		/// Returns index of photo with median exposure
		int GetMedianPhotoIndex();
		/// Returns input photo with a given index.
		std::shared_ptr <PhotoImage> GetInputPhoto(unsigned int index) {return this->inputPhotos[index];}
		std::shared_ptr <PhotoImage> GetMedianPhoto() {return this->inputPhotos[this->GetMedianPhotoIndex()];}
		std::vector <std::shared_ptr <PhotoImage> > GetInputPhotos() {return this->inputPhotos;}
		
		/// Returns result of processing.
		std::shared_ptr <PhotoImage> GetOutputPhoto() {return this->outputPhoto;}
		
		void SetNoiseReductionType(ENoiseReductionType type) {this->noiseReductionType = type;}
		/// Returns type of noise reduction.
		ENoiseReductionType GetNoiseReductionType() const {return this->noiseReductionType;}
		void SetNoiseReductionStrength(unsigned int strength) {this->noiseReductionStrength = strength;}
		
		void SetAlignmentEnabled(bool enabled) {this->alignImages = enabled;}
		bool IsAlignmentEnabled() const {return this->alignImages;}

		void SetAlignmentType(EAlignmentType type) {this->alignmentType = type;}
		EAlignmentType GetAlignmentType() const {return this->alignmentType;}

		void SetCropImagesEnabled(bool enabled) {this->cropImages = enabled;}
		bool IsCropEnabled() const {return this->cropImages;}

		void SetMaxShiftPerspective(float maxShift) {this->maxShiftPerspective = maxShift / 100.0f;}
		
		void SetColorTemperature(int temperature) {this->colorTemperature = temperature;}
		void SetColorSpace(unsigned int colorSpace) {this->colorSpace = colorSpace;}
		
		void SetBitsPerSampleOut(int bps) {this->bitsPerSampleOut = bps;}
		int GetBitsPerSampleOut() const {return this->bitsPerSampleOut;}
		
		void SetIsPreview(bool value) {this->isPreview = value;}
		bool IsPreview() const {return this->isPreview;}
		
		/** Returns output file name (combined from input photos file names).
		 * @param concatenate specifies the method the names will be combined. If it is true
		 * common part will be found, eg. for names: IMG_1, IMG_2, IMG_3 this will result in
		 * IMG_1_2_3 name. Otherwise name of the first file will be used.
		 */
		std::string GetOutputFileName(bool concatenate, bool includePath=true);

		/** Returns output file name (combined from input photos file names).
		 * @param concatenate specifies the method the names will be combined. If it is true
		 * common part will be found, eg. for names: IMG_1, IMG_2, IMG_3 this will result in
		 * IMG_1_2_3 name. Otherwise name of the first file will be used.
		 */
		static std::string GetOutputFileName(std::vector <std::string>& inputPhotos, bool concatenate);
		
		/// Specifies whether processor should load images before processing them.
		void SetLoadImages(bool value) {this->doLoadImages = value;}

		void SetPreview(bool value) {this->isPreview = value;}
		
		void SetDownsampleSize(unsigned int w, unsigned int h) {this->downsampleWidth = w; this->downsampleHeight = h;}
		
#ifdef WIN32
		static void __stdcall HandleProgress(float progress, int* cancel);
		void Start();

		void Stop() {stop = true;}
		bool IsStopped() {return stop;}

		// TODO: change this so it isn't static.
	#ifdef _CONSOLE
		// TODO: change this so it isn't static.
		static std::shared_ptr<ProgressNotifier> progressNotifier;
	#else
		static std::shared_ptr<Win32ProgressNotifier> progressNotifier;
	#endif
#else
		static void HandleProgress(float progress, int* cancel);

		// TODO: change this so it isn't static.
		static std::shared_ptr<ProgressNotifier> progressNotifier;
		
		void StartSync() {this->Process();}
#endif
	};
}


#endif // PHOTOPROCESSOR_H
